import copy
import json
import logging
import os
import re
import sys
import unittest
from contextlib import contextmanager

import responses
from pybsn.bcf.api import Api


pybsn_host = "http://127.0.0.1:8080"

my_dir = os.path.dirname(__file__)

logging.basicConfig()


@contextmanager
def responses_from_cassette(cassette):
    with responses.RequestsMock() as rsps:
        with open(os.path.join(my_dir, 'fixtures/cassettes/switch',cassette + ".json")) as file_:
            json_ = json.load(file_)

        def _handle_interaction_cb(req, interaction):
            if interaction["request"]["body"]["string"]:
                expected_req_body = json.loads(interaction["request"]["body"]["string"])
                assert json.loads(req.body) == expected_req_body

            return (interaction["response"]["status"]["code"], interaction["response"]["headers"],
                    interaction["response"]["body"]["string"])

        for h in json_["http_interactions"]:
            url = h["request"]["uri"].replace("<PYBSN_HOST>", pybsn_host)
            url = re.escape(url)
            url = url.replace(r'\[', r'(\[|%5B)').replace(r'\]', r'(\]|%5D)')
            #if True or sys.version_info < (3,0):
                # python2 hack - py2 quotes the square brackets, py3 does not
                #url = url.replace("[", "%5B").replace("]", "%5D")
            method = h["request"]["method"]
            cb = lambda req, interaction=h: _handle_interaction_cb(req, interaction)
            rsps.add_callback(method, re.compile(url),
                 callback = cb,
                 content_type = "application/json")

            print("registered callback %s %s" % (method, url))

        yield


class TestSwitch(unittest.TestCase):
    def setUp(self):
        self.api = Api('http://127.0.0.1:8080', "admin", "pass", login=False)

    def test_get_switches(self):
        with responses_from_cassette('get_switches'):
            switches = self.api.get_switches()

            assert len(switches) == 14

            for sw in switches:
                assert sw.name == None
                assert sw.dpid != None


    def test_get_switch_by_name(self):
        with responses_from_cassette('get_switch_by_name'):
            sw = self.api.get_switch_by_name("test")

            assert sw is not None
            assert sw.name == "test"

    def test_get_switch_by_name_fail(self):
        with responses_from_cassette('get_switch_by_name_fail'):
            sw = self.api.get_switch_by_name("test")

            assert sw is None

    def test_remove_switch_by_name(self):
        with responses_from_cassette('remove_switch_by_name'):
            # Add a switch first
            self.api.add_switch(dpid="00:00:00:00:01:01:01:01", name="test")

            # Verify it has been pushed
            sw = self.api.get_switch_by_name("test")
            assert sw is not None

            # Remove it
            self.api.remove_switch_by_name("test")

        with responses_from_cassette('remove_switch_by_name_2'):
            sw = self.api.get_switch_by_name("test")
            assert sw is None

    def test_remove_switch_by_dpid(self):
        with responses_from_cassette('remove_switch_by_dpid'):
            # Add a switch first
            self.api.add_switch(dpid="00:00:00:00:01:01:01:01", name="test")

            # Verify it has been puseed
            sw = self.api.get_switch_by_dpid("00:00:00:00:01:01:01:01")
            assert sw is not None

            # Remove it
            self.api.remove_switch_by_dpid("00:00:00:00:01:01:01:01")

        with responses_from_cassette('remove_switch_by_dpid_2'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:01:01:01:01")
            assert sw is None

    def test_get_switch_by_dpid(self):
        with responses_from_cassette('get_switch_by_dpid'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:00:01:00:01")

            assert sw is not None
            assert sw.dpid == "00:00:00:00:00:01:00:01"

    def test_get_switch_by_dpid_fail(self):
        with responses_from_cassette('get_switch_by_dpid_fail'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:00:00:00:01")

            assert sw is None

    def test_update(self):
        with responses_from_cassette('update'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:00:01:00:01")

            assert sw is not None

            sw.update()

            assert sw != None

    def test_remove(self):
        with responses_from_cassette("remove"):
            # Add a switch first
            self.api.add_switch(dpid="00:00:00:00:01:01:01:01", name="test")

            # Verify it has been puseed
            sw = self.api.get_switch_by_name("test")
            assert sw is not None

            # Remove it
            sw.remove()

        with responses_from_cassette("get_switch_by_name_fail"):
            sw = self.api.get_switch_by_name("test")
            assert sw is None


    def test_add_switch(self):
        with responses_from_cassette('add_switch'):
            sw = self.api.add_switch(dpid="00:00:00:00:01:01:01:01", name="test")

            assert sw is not None
            assert sw.name == "test"
            assert sw.dpid == "00:00:00:00:01:01:01:01"

            sw = self.api.get_switch_by_name("test")

            assert sw is not None


    def test_disconnect_switch(self):
        with responses_from_cassette('disconnect_switch'):
            switches = self.api.get_switches()

            assert len(switches) == 14

            for sw in switches:
                sw.disconnect()


    def test_get_interfaces(self):
        with responses_from_cassette('get_interfaces'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:00:01:00:01")

            assert sw is not None

            interfaces = sw.get_interfaces()

            assert len(interfaces) == 9


    def test_get_connections(self):
        with responses_from_cassette('get_connections'):
            sw = self.api.get_switch_by_dpid("00:00:00:00:00:01:00:01")

            assert sw is not None

            connections = sw.get_connections()

            assert len(connections) == 4



import json
import os
import logging
import re
import unittest

import requests

import pybsn
import responses

my_dir = os.path.dirname(__file__)

class TestBigDbClient(unittest.TestCase):
    def setUp(self):
        self.client = pybsn.connect("http://127.0.0.1:8080")

    @responses.activate
    def test_connect_login(self):
        def _login_cb(req):
            self.assertEqual(json.loads(req.body), {'user': 'admin', 'password': 'somepassword'})
            headers = {
                "Set-Cookie": "session_cookie=v_8yOI7FI-WKr-5QU6nxoJkVtAu5rKI-; Path=/api/;"
            }
            return (200, headers, json.dumps(
                {"success": True,
                 "session_cookie": "UPhNWlmDN0re8cg9xsqe9QT1QvQTznji",
                 "error_message": "",
                 "past_login_info":
                     {"failed_login_count": 0,
                      "last_success_login_info": {"host": "127.0.0.1", "timestamp": "2019-05-19T19:16:22.328Z"}}}))

        responses.add_callback(responses.POST, "http://127.0.0.1:8080/api/v1/auth/login", callback=_login_cb,
                               content_type="application/json")

        client = pybsn.connect("http://127.0.0.1:8080", "admin", "somepassword")

        # Responses is currently broken in that it doesn't retain the session cookies...
        # https://github.com/getsentry/responses/issues/80
        #
        # self.assertEqual(client.session.cookies, {"Cookie": "session_cookie=UPhNWlmDN0re8cg9xsqe9QT1QvQTznji"})
        #
        # def _get_cb(req):
        #     self.assertEqual(req.headers["Cookie"], "session_cookie=UPhNWlmDN0re8cg9xsqe9QT1QvQTznji")
        #     return (200, {}, json.dumps([ "ok" ]))
        #
        # responses.add_callback(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/test", callback=_get_cb,
        #                        content_type="application/json")
        # client.get("controller/test")

    @responses.activate
    def test_close_no_auth(self):
        self.client.close

    @responses.activate
    def test_connect_wrong_pw(self):
        def _login_cb(req):
            self.assertEqual(json.loads(req.body), {'user': 'admin', 'password': 'foo'})
            headers = {
                "Set-Cookie": "session_cookie=v_8yOI7FI-WKr-5QU6nxoJkVtAu5rKI-; Path=/api/;"
            }
            return (401, headers, json.dumps(
                {"success":False,
                 "session_cookie":None,
                 "error_message":"Invalid user/password combination.",
                 "past_login_info":None}))

        responses.add_callback(responses.POST, "http://127.0.0.1:8080/api/v1/auth/login", callback=_login_cb,
                               content_type="application/json")

        with self.assertRaises(requests.exceptions.HTTPError) as context:
            pybsn.connect("http://127.0.0.1:8080", "admin", "foo")
        self.assertEqual(context.exception.response.status_code, 401)

    @responses.activate
    def test_connect_token(self):
        def _aaa_cb(req):
            self.assertEqual(req.headers["Cookie"], "session_cookie=some_token")
            return (200, {}, json.dumps([ {  "auth-context-type" : "session-token",  "user-info" :
                              {   "full-name" : "Default admin",
                             "group" : [ "admin" ],    "user-name" : "admin"  } } ])
                    )

        responses.add_callback(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/core/aaa/auth-context",
                               callback=_aaa_cb, content_type="application/json")

        pybsn.connect("http://127.0.0.1:8080", token="some_token")


    @responses.activate
    def test_connect_token_custom_headers(self):
        def _aaa_cb(req):
            self.assertEqual(req.headers["Cookie"], "session_cookie=some_token")
            self.assertEqual(req.headers["X-Forwarded-For"], "1.2.3.4")
            return (200, {}, json.dumps([ {  "auth-context-type" : "session-token",  "user-info" :
                              {   "full-name" : "Default admin",
                             "group" : [ "admin" ],    "user-name" : "admin"  } } ])
                    )

        responses.add_callback(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/core/aaa/auth-context",
                               callback=_aaa_cb, content_type="application/json")

        pybsn.connect("http://127.0.0.1:8080", token="some_token", session_headers={"X-Forwarded-For": "1.2.3.4"})

    @responses.activate
    def test_connect_token_wrong(self):
        def _aaa_cb(req):
            self.assertEqual(req.headers["Cookie"], "session_cookie=wrong_token")
            return (401, {}, json.dumps({
                "description":"Authorization failed: No session or token found for cookie",
                "error-code":401
            }))

        responses.add_callback(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/core/aaa/auth-context",
                               callback=_aaa_cb, content_type="application/json")

        with self.assertRaises(requests.exceptions.HTTPError) as context:
            pybsn.connect("http://127.0.0.1:8080", token="wrong_token")
        self.assertEqual(context.exception.response.status_code, 401)

    @responses.activate
    def test_close_no_auth(self):
        # no request
        self.client.close()

    @responses.activate
    def test_close_session(self):
        responses.add(method=responses.DELETE,
                      url=re.compile(r"http://127\.0\.0\.1:8080/api/v1/data/controller/core/aaa/session"
                                     r"(?:%5B|\[)auth-token='value'(?:%5D|\])"),
                      status=204)

        self.client.session.cookies.set_cookie(
            requests.cookies.create_cookie(name="session_cookie", value="value"))
        self.client.close()

    @responses.activate
    def test_client_contextmanager(self):
        responses.add(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/core/healthy",
                      status=200, json= [ { "status" : "healthy"}] )

        with pybsn.connect("http://127.0.0.1:8080") as client:
            client.get("controller/core/healthy")

    @responses.activate
    def test_client_contextmanager_with_session(self):
        responses.add(method=responses.DELETE,
                      url=re.compile(r"http://127\.0\.0\.1:8080/api/v1/data/controller/core/aaa/session"
                                     r"(?:%5B|\[)auth-token='value'(?:%5D|\])"),
                      status=204)
        responses.add(responses.GET, "http://127.0.0.1:8080/api/v1/data/controller/core/healthy",
                      status=200, json= [ { "status" : "healthy"}] )

        with pybsn.connect("http://127.0.0.1:8080") as client:
            client.session.cookies.set_cookie(
                requests.cookies.create_cookie(name="session_cookie", value="value"))
            client.get("controller/core/healthy")

    @responses.activate
    def test_rpc(self):
        responses.add(responses.POST, "http://127.0.0.1:8080/api/v1/rpc/controller/rpc",
                      json={'id': 1234}, status=200)
        result = self.client.rpc(path="controller/rpc",
                                 data={"description": "desc"})
        self.assertEqual(result, {'id': 1234})

    @responses.activate
    def test_no_response(self):
        responses.add(responses.POST, "http://127.0.0.1:8080/api/v1/rpc/controller/rpc",
                      status=204)
        result = self.client.rpc(path="controller/rpc",
                                 data={"description": "desc"})
        self.assertIsNone(result)



import json
import logging
import re
from string import Template

import requests

try:
    import warnings
    from requests.packages.urllib3.exceptions import InsecureRequestWarning
    warnings.simplefilter("ignore", InsecureRequestWarning)
except ImportError:
    pass

logger = logging.getLogger("pybsn")

DATA_PREFIX = "/api/v1/data/"
RPC_PREFIX = "/api/v1/rpc/"
SCHEMA_PREFIX = "/api/v1/schema/"


# for python2, need to handle unicode strings specially in filter
try:
    UNICODE_TYPE = unicode
except NameError:
    UNICODE_TYPE = None

class Node(object):
    def __init__(self, path, connection):
        self._path = path
        self._connection = connection

    def __getattr__(self, name):
        return self[name.replace("_", "-")]

    def __getitem__(self, name):
        return Node(self._path + "/" + name, self._connection)

    def get(self, params=None):
        return self._connection.get(self._path, params)

    def post(self, data):
        return self._connection.post(self._path, data)

    def put(self, data):
        return self._connection.put(self._path, data)

    def patch(self, data):
        return self._connection.patch(self._path, data)

    def delete(self):
        return self._connection.delete(self._path)

    def schema(self):
        return self._connection.schema(self._path)

    def rpc(self, data=None):
        return self._connection.rpc(self._path, data)

    def filter(self, template, *args, **kwargs):
        # TODO escape values better than repr()
        def _convert(v):
            # fix for py2: unicode strings are encoded as u'foo' - we can't have that in a predicate
            # so convert to a "normal" string first.
            if UNICODE_TYPE and isinstance(v, UNICODE_TYPE):
                return repr(str(v))
            else:
                return repr(v)

        kwargs = { k: _convert(v) for k, v in kwargs.items() }
        predicate = '[' + Template(template).substitute(**kwargs) + ']'
        return Node(self._path + predicate, self._connection)

    def match(self, **kwargs):
        for k, v in kwargs.items():
            self = self.filter("%s=$x" % k.replace('_', '-'), x=v)
        return self

    def __call__(self):
        return self.get()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __repr__(self):
        return "Node(%s)" % self._path

class BigDbClient(object):
    def __init__(self, url, session):
        self.url = url
        self.session = session
        self.root = Node("controller", self)

    def __enter__(self):
        return self

    def __exit__(self, type, value, trace_back):
        self.close()

    def request(self, method, path, data=None, params=None, rpc=False):
        url = self.url + (RPC_PREFIX if rpc else DATA_PREFIX) + path

        request = requests.Request(method=method, url=url, data=data, params=params)
        response = logged_request(self.session, request)

        try:
            # Raise an HTTPError for 4xx/5xx codes
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if e.response.text:
                error_json = json.loads(e.response.text)
                # Attempt to capture the REST API error description and pass it along to the HTTPError
                if 'description' in error_json:
                    e.args = (e.args[0] + ': ' + error_json['description'],)
            raise
        return response

    def get(self, path, params=None):
        return self.request("GET", path, params=params).json()

    def rpc(self, path, data):
        response = self.request("POST", path, data=self._dump_if_present(data), rpc=True)
        if response.status_code == requests.codes.no_content:
            return None
        else:
            return response.json()

    def post(self, path, data):
        return self.request("POST", path, data=self._dump_if_present(data))

    def put(self, path, data):
        return self.request("PUT", path, data=self._dump_if_present(data))

    def patch(self, path, data):
        return self.request("PATCH", path, data=self._dump_if_present(data))

    def close(self):
        token = self.session.cookies.get_dict().get("session_cookie")
        if token:
            # This is a no-op/fine for api tokens
            self.root.core.aaa.session.match(auth_token=token).delete()

    def _dump_if_present(self, data):
        if data is not None:
            return json.dumps(data)
        else:
            return None

    def delete(self, path):
        return self.request("DELETE", path)

    def schema(self, path=""):
        url = self.url + SCHEMA_PREFIX + path
        response = self.session.get(url)
        response.raise_for_status()
        return json.loads(response.text)

    def __repr__(self):
        return "BigDbClient(%s)" % self.url


BIGDB_PROTO_PORTS = [
    ('https', 8443),
    ('http', 8080),
]


def logged_request(session, request):
    prepared = session.prepare_request(request)

    marker = "-" * 30
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("%s Request: %s\n%s\n%s\n\n%s", marker, marker, prepared.method + ' ' + prepared.url,
                  '\n'.join('{}: {}'.format(k, v) for k, v in prepared.headers.items()),
                  prepared.body)

    response = session.send(prepared)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("%s Response: %s\n%s\n%s\n\n%s", marker, marker, response.status_code,
                  '\n'.join('{}: {}'.format(k, v) for k, v in response.headers.items()),
                  response.content)

    return response


def guess_url(session, host, validate_path="/api/v1/auth/healthy"):
    if re.match(r'^https?://', host):
        return host
    else:
        for schema, port in BIGDB_PROTO_PORTS:
            url = "%s://%s:%d" % (schema, host, port)
            try:
                response = session.get(url + validate_path, timeout=2)
            except requests.exceptions.ConnectionError as e:
                logger.debug("Error connecting to %s: %s", url, str(e))
                continue
            if response.status_code == 200: # OK
                return url
            else:
                logger.debug("Could connect to URL %s: %s", url, response)
    raise Exception("Could not find available BigDB service on {}".format(host))

def attempt_login(session, url, username, password):
    auth_data = json.dumps({ 'user': username, 'password': password })
    path = "/api/v1/auth/login"
    request = requests.Request(method="POST", url=url + path, data=auth_data)
    response = logged_request(session, request)
    if response.status_code == 200: # OK
        # Fix up cookie path
        for cookie in session.cookies:
            if cookie.path == "/auth":
                cookie.path = "/api"
        return url
    else:
        response.raise_for_status()


def connect(host, username=None, password=None, token=None, login=None, verify_tls=False, session_headers=None):
    session = requests.Session()
    session.verify = verify_tls
    if session_headers:
        for k, v in session_headers.items():
            session.headers[k] = v

    url = guess_url(session, host)
    if login is None:
        login = (token is None) and username is not None and password is not None

    if login:
        attempt_login(session=session, url=url, username=username, password=password)
    elif token:
        cookie = requests.cookies.create_cookie(name="session_cookie", value=token)
        session.cookies.set_cookie(cookie)
        request = requests.Request(method="GET", url=url + "/api/v1/data/controller/core/aaa/auth-context")
        response = logged_request(session, request)
        if response.status_code != 200:
            response.raise_for_status()

    return BigDbClient(url, session)



from setuptools import setup

setup(name='pybsn',
    version='0.3.1',
    description="pybsn is a python interface to Big Switch Networks' products",
    url='https://github.com/floodlight/pybsn',
    author='Jason Parraga, Rich Lane, Andreas Wundsam',
    author_email='Sovietaced@gmail.com, Rich.Lane@bigswitch.com, Andreas.Wundsam@bigswitch.com',
    license='ECLIPSE',
    packages=['pybsn'],
    zip_safe=False,
    install_requires=[
    "requests >= 2.3.0"
    ],
    tests_require=[
        "responses >= 0.10.6"
    ],
    scripts=['bin/pybsn-repl', 'bin/pybsn-schema'],
    test_suite="test",
)
